@@ -1,16 +1,6 @@
-// generic datatables / Editor handling
-
-// data is an list of objects for rendering or url for ajax retrieval of similar object
-// buttons is a JSON parsable string, as it references editor which hasn't been instantiated yet
-// options is an object with the following keys
-//     dtopts:       options to be passed to DataTables instance, 
-//                   except for data: and buttons: options, passed in data, buttons
-//     editoropts:   options to be passed to Editor instance, 
-//                   if not present, Editor will not be configured
-//     updateopts:   options to configure Editor select fields with
-//                   see crudapi.py for more details
-//     yadcfopts:    yadcf options to be passed to yadcf 
-//                   if not present, yadcf will not be configured
+/**
+ * generic datatables / Editor handling
+ */
 
 var editor, _dt_table;
 var opttree = {};
@@ -70,12 +60,13 @@ function firstDataTableScrollAdjust() {
 }
 
 function jsDataTableScrollAdjust() {
-    if(_dt_table) {
-        var height = jsGetDataTableHeightPx() + "px";
-        $('.dataTables_scrollBody:has(#datatable)').css('max-height', height);
-        $('.DTFC_LeftBodyLiner').css('max-height', height);
-        _dt_table.draw();
-    }
+    return;
+    // if(_dt_table) {
+    //     var height = jsGetDataTableHeightPx() + "px";
+    //     $('.dataTables_scrollBody:has(#datatable)').css('max-height', height);
+    //     $('.DTFC_LeftBodyLiner').css('max-height', height);
+    //     _dt_table.draw();
+    // }
 }
 
 /**
@@ -116,6 +107,52 @@ function jsGetDataTableHeightPx() {
     return retHeightPx;
 }
 
+/**
+ * get the button options, correctly annotated with indicated editor
+ *
+ * @param buttons - list of editor actions 'create', 'edit', 'editRefresh', 'editChildRowRefresh', 'remove'
+ *          or other button configuration.
+ *          See https://datatables.net/reference/option/buttons and https://datatables.net/reference/option/buttons.buttons
+ * @param editor - editor to annotate action with
+ * @returns {[{extend:button, editor: editor}, {buttons.buttons fields}, ...]}
+ */
+function get_button_options(buttons, editor) {
+    var button_options = [];
+    for (i=0; i<buttons.length; i++) {
+        button = buttons[i];
+        if ($.inArray(button, ['create', 'edit', 'editRefresh', 'editChildRowRefresh', 'remove']) >= 0) {
+            button_options.push({extend:button, editor:editor});
+        } else {
+            // convert button actions to javascript, // kludge for conversion from python
+            if (button.hasOwnProperty('action')) {
+                button.action = eval(button.action)
+            }
+
+            button_options.push(button);
+        }
+    }
+
+    return button_options;
+}
+
+/**
+ * configure dataTables for table with id=datatable
+ *
+ * @param data - list of objects for rendering or url for ajax retrieval of similar object
+ * @param buttons - is a JSON parsable string, as it references editor which hasn't been instantiated yet
+ * @param options - object with the following keys
+ *     dtopts:       options to be passed to DataTables instance,
+ *                   except for data: and buttons: options, passed in data, buttons
+ *     editoropts:   options to be passed to Editor instance,
+ *                   if not present, Editor will not be configured
+ *     updateopts:   options to configure Editor select fields with
+ *                   see tables.py for more details
+ *     yadcfopts:    (optional) yadcf options to be passed to yadcf
+ *                   if not present, yadcf will not be configured
+ *     childrow:     (optional) options to configure ChildRow display (see datatables-childrow.js)
+ *                   if not present, childrow will not be configured
+ * @param files - (optional) passed to Editor instance
+ */
 function datatables(data, buttons, options, files) {
 
     // convert render to javascript -- backwards compatibility
@@ -152,10 +189,49 @@ function datatables(data, buttons, options, files) {
           }
         } );
 
+        // configure childrow options for editor if so configured
+        if ( ! $.isEmptyObject( options.childrow ) ) {
+            if (options.childrow.showeditor) {
+                $.extend(options.editoropts,{display:onPageDisplay('#childrow-editform-')})
+            }
+        }
+
         // create editor instance
         $.extend(options.editoropts,{table:'#datatable'})
         editor = new $.fn.dataTable.Editor ( options.editoropts );
 
+        // adapted from https://editor.datatables.net/examples/api/confirmClose
+        // requires Editor 1.9.5 (or patch to editor.jqueryui.js (editor.jqueryui.patch-discussion-63653.js)
+        var openVals;
+        editor
+            .on( 'open', function () {
+                // Store the values of the fields on open
+                openVals = JSON.stringify( editor.get() );
+
+                editor.on( 'preClose', function ( e ) {
+                    // On close, check if the values have changed and ask for closing confirmation if they have
+                    if ( openVals !== JSON.stringify( editor.get() ) ) {
+                        return confirm( 'You have unsaved changes. Are you sure you want to exit?' );
+                    }
+                } );
+            } )
+            .on( 'postCreate postEdit close', function () {
+                editor.off( 'preClose' );
+            } );
+
+        // if createfieldvals requested, add a handler which initializes fields when create form displayed
+        if (options.createfieldvals !== undefined) {
+            // save for initCreate function
+            editor.createfieldvals = options.createfieldvals;
+            editor.on('initCreate.dt', function(e) {
+                var that = this;
+                $.each(this.createfieldvals, function(field, val) {
+                    that.field(field).val(val);
+                });
+            });
+        }
+
+        // set up to update select options
         if (options.updateopts !== undefined) {
             for (i=0; i<options.updateopts.length; i++) {
                 updateopt = options.updateopts[i]
@@ -185,35 +261,41 @@ function datatables(data, buttons, options, files) {
     buttons = checkeval(buttons);
 
     // set up buttons, special care for editor buttons
-    var button_options = [];
-    for (i=0; i<buttons.length; i++) {
-        button = buttons[i];
-        if ($.inArray(button, ['create', 'edit', 'editRefresh', 'remove']) >= 0) {
-            button_options.push({extend:button, editor:editor});
-        } else {
-            // convert button actions to javascript, // kludge for conversion from python
-            if (button.hasOwnProperty('action')) {
-                button.action = eval(button.action)
-            }
-
-            button_options.push(button);
-        }
-    };
-
+    var button_options = get_button_options(buttons, editor);
     $.extend(options.dtopts, {buttons:button_options});
 
+    // handle rowReorder if requested; if no editor, disable rowReorder
+    // assume an object was configured
+    if (options.dtopts.rowReorder) {
+        if (options.editoropts) {
+            options.dtopts.rowReorder.editor = editor;
+            options.dtopts.rowReorder.update = false;
+            // 'row-reorder' event added later
+        } else {
+            options.dtopts.rowReorder = false;
+        }
+    }
+
     // assume data is url if serverSide is truthy
     if (options.dtopts.serverSide) {
-        $.extend(options.dtopts, { ajax: data });
+        var url = data;
+        // translate_datatable_group returns function which does the ajax data query, if doesn't exist, use url directly
+        $.extend(options.dtopts, { ajax: (translate_datatable_group && translate_datatable_group(url)) || url });
 
     // otherwise assume it is object containing the data to render
     } else {
         $.extend(options.dtopts, { data: data });
-    };
+    }
 
     // define the table
     _dt_table = $('#datatable').DataTable ( options.dtopts );
 
+    // configure childrow if so configured
+    if ( ! $.isEmptyObject( options.childrow ) ) {
+        // sets up child row event handling, and initializes child elements as needed
+        var childrow = new ChildRow(_dt_table, options.childrow, editor, childrow_childbase);
+    }
+
     // any column filtering required? if so, define the filters
     if ( ! $.isEmptyObject( options.yadcfopts ) ) {
         // general options supplied
@@ -231,10 +313,25 @@ function datatables(data, buttons, options, files) {
         }
     }
 
+    // handle rowReorder if requested; if no editor, disable rowReorder
+    // https://datatables.net/reference/event/row-reorder
+    if (options.dtopts.rowReorder) {
+        if (options.editoropts) {
+            // 'row-reorder' event added
+            _dt_table.on( 'row-reorder', function ( e, details, changes ) {
+                editor
+                    .edit( changes.nodes, false, {
+                        submit: 'allIfChanged'
+                    } )
+                    .multiSet( changes.dataSrc, changes.values )
+                    .submit();
+            });
+        }
+    }
     // take care of any initialization which needs to be done after datatables is initialized
     if (typeof afterdatatables !== "undefined") {
         afterdatatables();
-    };
+    }
 
     // adjust scrolling to fit window
     firstDataTableScrollAdjust();        
@@ -282,7 +379,7 @@ function EditorButtonDialog(options) {
             this.open()
         } else {
             this.close()
-        };
+        }
     };
 
     this.open = function() {
